import logging
import re

from AcdhArcheAssets.uri_norm_rules import get_normalized_uri
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.db.models import BooleanField, CharField, TextField
from django.db.models.fields.related import ForeignKey, ManyToManyField
from django.db.models.query import QuerySet
from django.forms import model_to_dict
from django.urls import reverse

from apis_core.generic.helpers import mro_paths, permission_fullname
from apis_core.generic.signals import (
    post_duplicate,
    post_merge_with,
    pre_duplicate,
    pre_import_from,
    pre_merge_with,
)
from apis_core.utils.settings import apis_base_uri, rdf_namespace_prefix

logger = logging.getLogger(__name__)


class GenericModel(models.Model):
    class Meta:
        abstract = True

    def __repr__(self):
        if id := getattr(self, "id", None):
            return super().__repr__() + f" (ID: {id})"
        return super().__repr__()

    @property
    def content_type(self):
        return ContentType.objects.get_for_model(self)

    @classmethod
    def get_listview_url(cls):
        ct = ContentType.objects.get_for_model(cls)
        return reverse("apis_core:generic:list", args=[ct])

    @classmethod
    def get_createview_url(cls):
        ct = ContentType.objects.get_for_model(cls)
        return reverse("apis_core:generic:create", args=[ct])

    @classmethod
    def get_importview_url(cls):
        ct = ContentType.objects.get_for_model(cls)
        return reverse("apis_core:generic:import", args=[ct])

    @classmethod
    def get_openapi_tags(cls):
        return [item[-1] for item in mro_paths(cls)]

    @classmethod
    def get_namespace_prefix(cls):
        ct = ContentType.objects.get_for_model(cls)
        return f"{rdf_namespace_prefix()}-{ct.model}"

    @classmethod
    def get_namespace_uri(cls):
        return apis_base_uri() + cls.get_listview_url()

    @classmethod
    def get_rdf_types(cls):
        return []

    def get_edit_url(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:generic:update", args=[ct, self.id])

    def get_duplicate_url(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:generic:duplicate", args=[ct, self.id])

    def get_enrich_url(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:generic:enrich", args=[ct, self.id])

    def get_absolute_url(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:generic:detail", args=[ct, self.id])

    def get_delete_url(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:generic:delete", args=[ct, self.id])

    def get_merge_url(self, other_id):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:generic:merge", args=[ct, self.id, other_id])

    def get_select_merge_or_enrich_url(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:generic:selectmergeorenrich", args=[ct, self.id])

    def get_create_success_url(self):
        return self.get_absolute_url()

    def get_update_success_url(self):
        return self.get_edit_url()

    def get_api_detail_endpoint(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:generic:genericmodelapi-detail", args=[ct, self.id])

    @classmethod
    def get_change_permission(self):
        return permission_fullname("change", self)

    @classmethod
    def get_add_permission(self):
        return permission_fullname("add", self)

    @classmethod
    def get_delete_permission(self):
        return permission_fullname("delete", self)

    @classmethod
    def get_view_permission(self):
        return permission_fullname("view", self)

    @classmethod
    def get_verbose_name_plural(cls):
        return cls._meta.verbose_name_plural

    @classmethod
    def get_verbose_name(cls):
        return cls._meta.verbose_name

    @classmethod
    def valid_import_url(cls, uri: str):
        """
        Check if an URI is a can be imported.
        The exact fetching logic for an URI is defined in the
        `import_definitions` attribute of the class.
        `import_definitions` has to be a dict, mapping a regex
        matching the URI to a callable taking the URI as an argument.
        This method check if there is a callable defined for this URI.
        """
        uri = get_normalized_uri(uri)
        for regex, fn in getattr(cls, "import_definitions", {}).items():
            if re.match(regex, uri):
                return fn
        return False

    @classmethod
    def fetch_from(cls, uri: str):
        """
        Fetch data from an URI. Check if there is import logic
        configured for this URI and if so, use that import logic
        to fetch the data.
        """
        uri = get_normalized_uri(uri)
        if fn := cls.valid_import_url(uri):
            return fn(uri) or {}
        raise ImproperlyConfigured(f"Import not configured for URI {uri}")

    @classmethod
    def import_from(cls, uri: str, allow_empty: bool = True):
        """
        Fetch data from an URI and create a model instance using
        that data. If the `allow_empty` argument is set, this also
        creates a model instance if the data fetched was empty. This
        might make sense if you still want to create an instance and
        attach the URI to it.
        """
        uri = get_normalized_uri(uri)
        # we allow other apps to injercept the import
        # whatever they return will be used instead of
        # creating a new object
        for receiver, response in pre_import_from.send(sender=cls, uri=uri):
            if response:
                return response
        data = cls.fetch_from(uri) or {}
        if allow_empty or data:
            instance = cls()
            instance.import_data(data)
            instance._uris = [data.get("uri", uri)]
            instance.save()
            return instance
        raise ValueError(f"Could not fetch data to import from {uri}")

    def import_from_dict_subset(self, **data) -> dict:
        """
        Import attributes of this instance from data in a dict.
        We iterate through the individual values of the dict and
        a) only set them if the instance has an attribute matching
        the key and b) use the fields `clean` method to check if
        the value validates. If it does not validate, we return
        the validation error in the errors dict.
        """
        errors = {}
        if data:
            for field in self._meta.fields:
                if data.get(field.name, False):
                    value = str(data[field.name][0])
                    try:
                        field.clean(value, self)
                    except Exception as e:
                        logger.info(
                            "Could not set %s on %s: %s", field.name, str(self), str(e)
                        )
                        errors[field.name] = str(e)
                    else:
                        setattr(self, field.name, value)
            self.save()
        return errors

    def import_data(self, data) -> dict:
        return self.import_from_dict_subset(**data)

    def get_merge_charfield_value(self, other: CharField, field: CharField):
        res = getattr(self, field.name)
        if not field.choices:
            otherres = getattr(other, field.name, res)
            if otherres and otherres != res:
                res += f" ({otherres})"
        return res

    def get_merge_textfield_value(self, other: TextField, field: TextField):
        res = getattr(self, field.name)
        if getattr(other, field.name):
            # if own value is None, fallback to empty string
            res = res or ""
            res += "\n" + f"Merged from {other}:\n" + getattr(other, field.name)
        return res

    def get_merge_booleanfield(self, other: BooleanField, field: BooleanField):
        return getattr(other, field.name)

    def get_field_value_after_merge(self, other, field):
        """
        This method finds the value of a field after merging `other` into `self`.
        It first tries to find a merge method that is specific to that field
        (merge_{fieldname}) and then tries to find a method that is specific to
        the type of the field (merge_{fieldtype})
        If neither of those exist, it uses the others field value if the field
        in self is not set, otherwise it keeps the value in self.
        """
        fieldtype = field.get_internal_type().lower()
        # if there is a `get_merge_{fieldname}` method in this model, use that one
        if callable(getattr(self, f"get_merge_{field.name}_value", None)):
            return getattr(self, f"get_merge_{field.name}_value")(other)
        # otherwise we check if there is a method for the field type and use that one
        elif callable(getattr(self, f"get_merge_{fieldtype}_value", None)):
            return getattr(self, f"get_merge_{fieldtype}_value")(other, field)
        else:
            if not getattr(self, field.name):
                return getattr(other, field.name)
        return getattr(self, field.name)

    def merge_fields(self, other):
        """
        This method iterates through the model fields and uses the
        `get_field_value_after_merge` method to copy values from `other` to `self`.
        It is called by the `merge_with` method.
        """
        for field in self._meta.fields:
            newval = self.get_field_value_after_merge(other, field)
            if newval != getattr(self, field.name):
                setattr(self, field.name, newval)
        self.save()

    def merge_with(self, entities):
        if self in entities:
            entities.remove(self)
        origin = self.__class__
        pre_merge_with.send(sender=origin, instance=self, entities=entities)

        e_a = type(self).__name__
        self_model_class = ContentType.objects.get(model__iexact=e_a).model_class()
        if isinstance(entities, int):
            entities = self_model_class.objects.get(pk=entities)
        if not isinstance(entities, list) and not isinstance(entities, QuerySet):
            entities = [entities]
            entities = [
                self_model_class.objects.get(pk=ent) if isinstance(ent, int) else ent
                for ent in entities
            ]
        for ent in entities:
            e_b = type(ent).__name__
            if e_a != e_b:
                continue
            for f in ent._meta.local_many_to_many:
                if not f.name.endswith("_set"):
                    sl = list(getattr(self, f.name).all())
                    for s in getattr(ent, f.name).all():
                        if s not in sl:
                            getattr(self, f.name).add(s)

        for ent in entities:
            self.merge_fields(ent)

        post_merge_with.send(sender=origin, instance=self, entities=entities)

        for ent in entities:
            ent.delete()

    def duplicate(self):
        origin = self.__class__
        pre_duplicate.send(sender=origin, instance=self)
        # usually, copying instances would work like
        # https://docs.djangoproject.com/en/4.2/topics/db/queries/#copying-model-instances
        # but we are working with abstract classes,
        # so we have to do it by hand  using model_to_dict:(
        objdict = model_to_dict(self)

        # remove unique fields from dict representation
        unique_fields = [field for field in self._meta.fields if field.unique]
        for field in unique_fields:
            logger.info(f"Duplicating {self}: ignoring unique field {field.name}")
            objdict.pop(field.name, None)

        # remove related fields from dict representation
        related_fields = [
            field for field in self._meta.get_fields() if field.is_relation
        ]
        for field in related_fields:
            objdict.pop(field.name, None)

        newobj = type(self).objects.create(**objdict)

        for field in related_fields:
            # we are not using `isinstance` because we want to
            # differentiate between different levels of inheritance
            if type(field) is ForeignKey:
                setattr(newobj, field.name, getattr(self, field.name))
            if type(field) is ManyToManyField:
                objfield = getattr(newobj, field.name)
                values = getattr(self, field.name).all()
                objfield.set(values)

        newobj.save()
        post_duplicate.send(sender=origin, instance=self, duplicate=newobj)
        return newobj

    duplicate.alters_data = True

    def uri_set(self):
        ct = ContentType.objects.get_for_model(self)
        return (
            ContentType.objects.get(app_label="uris", model="uri")
            .model_class()
            .objects.filter(content_type=ct, object_id=self.id)
            .all()
        )

    def uri_set_with_importer(self):
        return [uri for uri in self.uri_set() if self.valid_import_url(uri.uri)]



from django.db import models

from apis_core.history.models import VersionMixin
from apis_core.relations.models import Relation


class Profession(VersionMixin, models.Model):
    label = models.CharField()


class Person(VersionMixin, models.Model):
    first_name = models.CharField()
    last_name = models.CharField()
    profession = models.ManyToManyField(Profession, blank=True)

    def __str__(self):
        return f"{self.first_name} {self.last_name}"


class Place(VersionMixin, models.Model):
    label = models.CharField()


class PersonPlaceRelation(VersionMixin, Relation):
    subj_model = Person
    obj_model = Place



from django.contrib import admin

from apis_core.collections.models import SkosCollection, SkosCollectionContentObject

admin.site.register(SkosCollection)
admin.site.register(SkosCollectionContentObject)



from drf_spectacular.utils import extend_schema_field
from rest_framework.serializers import SerializerMethodField

from apis_core.generic.serializers import (
    GenericHyperlinkedModelSerializer,
    SimpleObjectSerializer,
    serializer_factory,
)


class RelationSerializer(GenericHyperlinkedModelSerializer):
    subj = SerializerMethodField()
    obj = SerializerMethodField()

    class Meta:
        exclude = (
            "subj_content_type",
            "subj_object_id",
            "obj_content_type",
            "obj_object_id",
        )

    @extend_schema_field(SimpleObjectSerializer())
    def get_subj(self, obj):
        if obj.subj:
            serializer = serializer_factory(type(obj.subj), SimpleObjectSerializer)
            return serializer(
                obj.subj, context={"request": self.context["request"]}
            ).data

    @extend_schema_field(SimpleObjectSerializer())
    def get_obj(self, obj):
        if obj.obj:
            serializer = serializer_factory(type(obj.obj), SimpleObjectSerializer)
            return serializer(
                obj.obj, context={"request": self.context["request"]}
            ).data



from django import template
from django.conf import settings

from apis_core import __version__

register = template.Library()


@register.simple_tag
def shared_url():
    return getattr(settings, "SHARED_URL", "/static/")


@register.simple_tag
def page_range(paginator, number):
    return paginator.get_elided_page_range(number=number)


@register.filter
def opts(obj):
    return obj._meta


@register.filter
def model_meta(content_type, field):
    return getattr(content_type.model_class()._meta, field)


@register.simple_tag
def apis_version():
    return __version__


@register.simple_tag
def git_repository_url():
    return getattr(
        settings, "GIT_REPOSITORY_URL", "https://github.com/acdh-oeaw/apis-core-rdf/"
    )


@register.simple_tag
def get_model_fields(
    model, include_parents=True, include_hidden=False, exclude_parent_links=True
):
    """
    Return all fields from a model.
    This uses the Django built in `get_fields` method but also allows to exlude parent links.
    Parent links are the `_ptr` fields that point to a parent class.
    """
    fields = model._meta.get_fields(include_parents, include_hidden)
    if exclude_parent_links:
        parent_links = model._meta.parents.values()
        fields = [field for field in fields if field not in parent_links]
    return fields



# Generated by Django 5.0.4 on 2024-05-14 08:14

import django.db.models.deletion
import simple_history.models
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("apis_metainfo", "0012_remove_rootobject_deprecated_name"),
        ("contenttypes", "0002_remove_content_type_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Group",
            fields=[
                (
                    "rootobject_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="apis_metainfo.rootobject",
                    ),
                ),
                ("label", models.CharField(blank=True, default="", max_length=4096)),
            ],
            options={
                "abstract": False,
            },
            bases=("apis_metainfo.rootobject", models.Model),
        ),
        migrations.CreateModel(
            name="Place",
            fields=[
                (
                    "rootobject_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="apis_metainfo.rootobject",
                    ),
                ),
                ("label", models.CharField(blank=True, default="", max_length=4096)),
                ("longitude", models.FloatField(blank=True, null=True)),
                ("latitude", models.FloatField(blank=True, null=True)),
            ],
            options={
                "abstract": False,
            },
            bases=("apis_metainfo.rootobject", models.Model),
        ),
        migrations.CreateModel(
            name="Profession",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(blank=True, default="", max_length=1024)),
            ],
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Person",
            fields=[
                (
                    "rootobject_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="apis_metainfo.rootobject",
                    ),
                ),
                ("forename", models.CharField(blank=True, default="", max_length=4096)),
                ("surname", models.CharField(blank=True, default="", max_length=4096)),
                ("gender", models.CharField(blank=True, default="", max_length=4096)),
                ("date_of_birth", models.DateField(blank=True, null=True)),
                ("date_of_death", models.DateField(blank=True, null=True)),
                (
                    "profession",
                    models.ManyToManyField(blank=True, to="sample_project.profession"),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("apis_metainfo.rootobject", models.Model),
        ),
        migrations.CreateModel(
            name="VersionGroup",
            fields=[
                (
                    "rootobject_ptr",
                    models.ForeignKey(
                        auto_created=True,
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        parent_link=True,
                        related_name="+",
                        to="apis_metainfo.rootobject",
                    ),
                ),
                (
                    "id",
                    models.IntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                (
                    "version_tag",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                ("label", models.CharField(blank=True, default="", max_length=4096)),
                ("history_id", models.AutoField(primary_key=True, serialize=False)),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "self_contenttype",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name": "Version",
                "verbose_name_plural": "Versions",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(
                simple_history.models.HistoricalChanges,
                models.Model,
            ),
        ),
        migrations.CreateModel(
            name="VersionPerson",
            fields=[
                (
                    "rootobject_ptr",
                    models.ForeignKey(
                        auto_created=True,
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        parent_link=True,
                        related_name="+",
                        to="apis_metainfo.rootobject",
                    ),
                ),
                (
                    "id",
                    models.IntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                (
                    "version_tag",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                ("forename", models.CharField(blank=True, default="", max_length=4096)),
                ("surname", models.CharField(blank=True, default="", max_length=4096)),
                ("gender", models.CharField(blank=True, default="", max_length=4096)),
                ("date_of_birth", models.DateField(blank=True, null=True)),
                ("date_of_death", models.DateField(blank=True, null=True)),
                ("history_id", models.AutoField(primary_key=True, serialize=False)),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "self_contenttype",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name": "Version",
                "verbose_name_plural": "Versions",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(
                simple_history.models.HistoricalChanges,
                models.Model,
            ),
        ),
        migrations.CreateModel(
            name="VersionPerson_profession",
            fields=[
                (
                    "id",
                    models.BigIntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                ("m2m_history_id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "history",
                    models.ForeignKey(
                        db_constraint=False,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        to="sample_project.versionperson",
                    ),
                ),
                (
                    "person",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        db_tablespace="",
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="sample_project.person",
                    ),
                ),
                (
                    "profession",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        db_tablespace="",
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="sample_project.profession",
                    ),
                ),
            ],
            options={
                "verbose_name": "VersionPerson_profession",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="VersionPlace",
            fields=[
                (
                    "rootobject_ptr",
                    models.ForeignKey(
                        auto_created=True,
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        parent_link=True,
                        related_name="+",
                        to="apis_metainfo.rootobject",
                    ),
                ),
                (
                    "id",
                    models.IntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                (
                    "version_tag",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                ("label", models.CharField(blank=True, default="", max_length=4096)),
                ("longitude", models.FloatField(blank=True, null=True)),
                ("latitude", models.FloatField(blank=True, null=True)),
                ("history_id", models.AutoField(primary_key=True, serialize=False)),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "self_contenttype",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name": "Version",
                "verbose_name_plural": "Versions",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(
                simple_history.models.HistoricalChanges,
                models.Model,
            ),
        ),
    ]



from datetime import datetime

import pytest
from faker import Faker

from apis_core.history.tests.models import (
    Person,
    PersonPlaceRelation,
    Place,
    Profession,
)

fake = Faker()


@pytest.mark.django_db
class TestHistory(object):
    def test_history_date_setter(self):
        """Test the date setter of the VersionMixin."""
        person = Person.objects.create()
        person._history_date = datetime(1970, 1, 1, 0, 0).astimezone()
        person.save()
        assert (
            person.history.last().history_date
            == datetime(1970, 1, 1, 0, 0).astimezone()
        )

    def test_history(self):
        """Tests the simple version of attributes changed on a models instance."""
        first_name = fake.first_name()
        last_name = fake.last_name()
        person = Person.objects.create(first_name=first_name, last_name=last_name)
        new_first_name = fake.first_name()
        person.first_name = new_first_name
        person.save()
        assert person.first_name == new_first_name
        person_history = person.history.all()
        assert len(person_history) == 2
        assert person.history.most_recent().first_name == new_first_name
        assert person.history.earliest().first_name == first_name

    def test_history_delete_entry(self):
        """Tests the deletion of an entry."""
        person = Person.objects.create()
        person.delete()
        assert len(Person.history.all()) == 2

    def test_m2m_save(self):
        """Test if m2m profession is saved correctly."""
        person = Person.objects.create()
        person_history = person.history.all()
        assert len(person_history) == 1
        profession = Profession.objects.create(label=fake.job())
        person.profession.add(profession)
        person.save()
        assert len(person.profession.all()) == 1
        assert len(person.history.latest().profession.all()) == 1
        assert len(person.history.earliest().profession.all()) == 0

    def test_history_date(self):
        """Test that history is set correctly when not manually set."""
        person = Person.objects.create()
        person.first_name = fake.first_name()
        person.save()
        assert (
            person.history.earliest().history_date
            < person.history.latest().history_date
        )

    def test_relation_list(self):
        person = Person.objects.create()
        place = Place.objects.create()
        assert len(person.get_history_data()) == 1
        ppr = PersonPlaceRelation.objects.create_between_instances(person, place)
        assert len(person.get_history_data()) == 2
        place = Place.objects.create()
        ppr.obj = place
        ppr.save()
        assert len(person.get_history_data()) == 3



from apis_core.generic.tables import GenericTable, MoreLessColumn


class PersonTable(GenericTable):
    class Meta(GenericTable.Meta):
        sequence = ["desc", "bio"]

    # example column to show preview and detail
    bio = MoreLessColumn(
        orderable=False,
        preview=lambda x: f"This is {x.surname}",
        fulltext=lambda x: f"This is {x.surname}, {x.forename} {x.surname}.",
    )



import logging

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from apis_core.generic.signals import post_merge_with, pre_import_from
from apis_core.uris.models import Uri

logger = logging.getLogger(__name__)


@receiver(post_delete)
def remove_stale_uris(sender, instance, *args, **kwargs):
    content_type = ContentType.objects.get_for_model(instance)
    if isinstance(instance.pk, int):
        uris = Uri.objects.filter(content_type=content_type, object_id=instance.pk)
        for uri in uris:
            logger.info(
                "Deleting uri %s as a result of deleting %s", repr(uri), repr(instance)
            )
            uri.delete()


@receiver(post_save, dispatch_uid="create_default_uri")
def create_default_uri(sender, instance, created, raw, using, update_fields, **kwargs):
    # disable the handler during fixture loading
    if raw:
        return
    # The list of Uris that should be created
    uris = getattr(instance, "_uris", [])

    # If this is a new object (created) we ask the model for its
    # default Uri and add that to the list of Uris
    skip_default_uri = getattr(instance, "skip_default_uri", False)
    create_default_uri = getattr(settings, "CREATE_DEFAULT_URI", True)
    if created and create_default_uri and not skip_default_uri:
        try:
            uris.append(instance.get_default_uri())
        except AttributeError:
            pass

    # We first check if there are even any Uris to create before we
    # lookup the content_type. This is a bit of a workaround, because
    # during migration the signal is triggered, but the lookup would
    # fail.
    if uris:
        content_type = ContentType.objects.get_for_model(instance)
        for uri in uris:
            logger.info(
                "Creating uri %s as a result of saving %s", repr(uri), repr(instance)
            )
            Uri.objects.get_or_create(
                uri=uri, content_type=content_type, object_id=instance.id
            )


@receiver(post_merge_with)
def merge_uris(sender, instance, entities, *args, **kwargs):
    instance_content_type = ContentType.objects.get_for_model(instance)
    for entity in entities:
        content_type = ContentType.objects.get_for_model(entity)
        for uri in Uri.objects.filter(content_type=content_type, object_id=entity.id):
            logger.info(
                "Updating uri %s to point to %s as a result of merging %s",
                repr(uri),
                repr(instance),
                repr(entity),
            )
            uri.content_type = instance_content_type
            uri.object_id = instance.id
            uri.save()


@receiver(pre_import_from)
def import_from(sender, uri, **kwargs):
    try:
        uri = Uri.objects.get(uri=uri)
        return uri.content_object
    except Uri.DoesNotExist:
        pass



# Generated by Django 5.2.5 on 2025-09-03 06:55

import django.db.models.deletion
import simple_history.models
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    replaces = [
        ("sample_project", "0001_initial"),
        ("sample_project", "0002_iscousinof_ispartof_issiblingof_livesin"),
        ("sample_project", "0003_alter_versiongroup_options_and_more"),
        ("sample_project", "0004_remove_versiongroup_self_contenttype_and_more"),
        ("sample_project", "0005_place_feature_code_versionplace_feature_code"),
        (
            "sample_project",
            "0006_alter_group_label_alter_person_date_of_birth_and_more",
        ),
        ("sample_project", "0007_alter_person_forename_alter_versionperson_forename"),
        ("sample_project", "0008_remove_versiongroup_version_tag_and_more"),
        ("sample_project", "0009_versionprofession"),
        ("sample_project", "0010_alter_group_options_alter_person_options_and_more"),
        ("sample_project", "0011_alter_place_feature_code_and_more"),
    ]

    initial = True

    dependencies = [
        ("apis_metainfo", "0012_remove_rootobject_deprecated_name"),
        ("contenttypes", "0002_remove_content_type_name"),
        ("relations", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Profession",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(blank=True, default="", max_length=1024)),
            ],
        ),
        migrations.CreateModel(
            name="Person",
            fields=[
                (
                    "rootobject_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="apis_metainfo.rootobject",
                    ),
                ),
                (
                    "forename",
                    models.CharField(
                        blank=True, default="", max_length=4096, verbose_name="forename"
                    ),
                ),
                (
                    "surname",
                    models.CharField(
                        blank=True, default="", max_length=4096, verbose_name="surname"
                    ),
                ),
                (
                    "gender",
                    models.CharField(
                        blank=True, default="", max_length=4096, verbose_name="gender"
                    ),
                ),
                (
                    "date_of_birth",
                    models.DateField(
                        blank=True, null=True, verbose_name="date of birth"
                    ),
                ),
                (
                    "date_of_death",
                    models.DateField(
                        blank=True, null=True, verbose_name="date of death"
                    ),
                ),
                (
                    "profession",
                    models.ManyToManyField(blank=True, to="sample_project.profession"),
                ),
            ],
            options={
                "abstract": False,
                "ordering": ["surname", "forename"],
                "verbose_name": "person",
                "verbose_name_plural": "persons",
            },
            bases=("apis_metainfo.rootobject", models.Model),
        ),
        migrations.CreateModel(
            name="VersionPerson",
            fields=[
                (
                    "rootobject_ptr",
                    models.ForeignKey(
                        auto_created=True,
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        parent_link=True,
                        related_name="+",
                        to="apis_metainfo.rootobject",
                    ),
                ),
                (
                    "id",
                    models.IntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                (
                    "forename",
                    models.CharField(
                        blank=True, default="", max_length=4096, verbose_name="forename"
                    ),
                ),
                (
                    "surname",
                    models.CharField(
                        blank=True, default="", max_length=4096, verbose_name="surname"
                    ),
                ),
                (
                    "gender",
                    models.CharField(
                        blank=True, default="", max_length=4096, verbose_name="gender"
                    ),
                ),
                (
                    "date_of_birth",
                    models.DateField(
                        blank=True, null=True, verbose_name="date of birth"
                    ),
                ),
                (
                    "date_of_death",
                    models.DateField(
                        blank=True, null=True, verbose_name="date of death"
                    ),
                ),
                ("history_id", models.AutoField(primary_key=True, serialize=False)),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical person",
                "verbose_name_plural": "historical persons",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="VersionPerson_profession",
            fields=[
                (
                    "id",
                    models.BigIntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                ("m2m_history_id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "history",
                    models.ForeignKey(
                        db_constraint=False,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        to="sample_project.versionperson",
                    ),
                ),
                (
                    "person",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        db_tablespace="",
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="sample_project.person",
                    ),
                ),
                (
                    "profession",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        db_tablespace="",
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="sample_project.profession",
                    ),
                ),
            ],
            options={
                "verbose_name": "VersionPerson_profession",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="IsCousinOf",
            fields=[
                (
                    "relation_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="relations.relation",
                    ),
                ),
            ],
            bases=("relations.relation",),
        ),
        migrations.CreateModel(
            name="IsPartOf",
            fields=[
                (
                    "relation_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="relations.relation",
                    ),
                ),
            ],
            bases=("relations.relation",),
        ),
        migrations.CreateModel(
            name="IsSiblingOf",
            fields=[
                (
                    "relation_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="relations.relation",
                    ),
                ),
            ],
            bases=("relations.relation",),
        ),
        migrations.CreateModel(
            name="LivesIn",
            fields=[
                (
                    "relation_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="relations.relation",
                    ),
                ),
            ],
            bases=("relations.relation",),
        ),
        migrations.CreateModel(
            name="VersionGroup",
            fields=[
                (
                    "rootobject_ptr",
                    models.ForeignKey(
                        auto_created=True,
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        parent_link=True,
                        related_name="+",
                        to="apis_metainfo.rootobject",
                    ),
                ),
                (
                    "id",
                    models.IntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                (
                    "label",
                    models.CharField(
                        blank=True, default="", max_length=4096, verbose_name="label"
                    ),
                ),
                ("history_id", models.AutoField(primary_key=True, serialize=False)),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical group",
                "verbose_name_plural": "historical groups",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="VersionProfession",
            fields=[
                (
                    "id",
                    models.BigIntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                ("name", models.CharField(blank=True, default="", max_length=1024)),
                ("history_id", models.AutoField(primary_key=True, serialize=False)),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical profession",
                "verbose_name_plural": "historical professions",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="Group",
            fields=[
                (
                    "rootobject_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="apis_metainfo.rootobject",
                    ),
                ),
                (
                    "label",
                    models.CharField(
                        blank=True, default="", max_length=4096, verbose_name="label"
                    ),
                ),
            ],
            options={
                "abstract": False,
                "ordering": ["label"],
                "verbose_name": "group",
                "verbose_name_plural": "groups",
            },
            bases=("apis_metainfo.rootobject", models.Model),
        ),
        migrations.CreateModel(
            name="Place",
            fields=[
                (
                    "rootobject_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="apis_metainfo.rootobject",
                    ),
                ),
                (
                    "label",
                    models.CharField(
                        blank=True, default="", max_length=4096, verbose_name="label"
                    ),
                ),
                (
                    "longitude",
                    models.FloatField(blank=True, null=True, verbose_name="longitude"),
                ),
                (
                    "latitude",
                    models.FloatField(blank=True, null=True, verbose_name="latitude"),
                ),
                (
                    "feature_code",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text='<a href="https://www.geonames.org/export/codes.html" target="_blank">Geonames Feature Code List</a>',
                        max_length=16,
                        verbose_name="feature code",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "ordering": ["label"],
                "verbose_name": "place",
                "verbose_name_plural": "places",
            },
            bases=("apis_metainfo.rootobject", models.Model),
        ),
        migrations.CreateModel(
            name="VersionPlace",
            fields=[
                (
                    "rootobject_ptr",
                    models.ForeignKey(
                        auto_created=True,
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        parent_link=True,
                        related_name="+",
                        to="apis_metainfo.rootobject",
                    ),
                ),
                (
                    "id",
                    models.IntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                (
                    "label",
                    models.CharField(
                        blank=True, default="", max_length=4096, verbose_name="label"
                    ),
                ),
                (
                    "longitude",
                    models.FloatField(blank=True, null=True, verbose_name="longitude"),
                ),
                (
                    "latitude",
                    models.FloatField(blank=True, null=True, verbose_name="latitude"),
                ),
                ("history_id", models.AutoField(primary_key=True, serialize=False)),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "feature_code",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text='<a href="https://www.geonames.org/export/codes.html" target="_blank">Geonames Feature Code List</a>',
                        max_length=16,
                        verbose_name="feature code",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical place",
                "verbose_name_plural": "historical places",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
    ]



import functools
import logging

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.base import ModelBase
from model_utils.managers import InheritanceManager

from apis_core.generic.abc import GenericModel

logger = logging.getLogger(__name__)


class RelationManager(InheritanceManager):
    def create_between_instances(self, subj, obj, *args, **kwargs):
        subj_object_id = subj.pk
        subj_content_type = ContentType.objects.get_for_model(subj)
        obj_object_id = obj.pk
        obj_content_type = ContentType.objects.get_for_model(obj)
        rel = self.create(
            subj_object_id=subj_object_id,
            subj_content_type=subj_content_type,
            obj_object_id=obj_object_id,
            obj_content_type=obj_content_type,
        )
        logger.debug("Created relation %s between %s and %s", rel.name(), subj, obj)
        return rel


# This ModelBase is simply there to check if the needed attributes
# are set in the Relation child classes.
class RelationModelBase(ModelBase):
    def __new__(metacls, name, bases, attrs):
        if name == "Relation":
            return super().__new__(metacls, name, bases, attrs)
        else:
            new_class = super().__new__(metacls, name, bases, attrs)
            if not (new_class._meta.abstract or new_class._meta.proxy):
                if not hasattr(new_class, "subj_model"):
                    raise ValueError(
                        "%s inherits from Relation and must therefore specify subj_model"
                        % name
                    )
                if not hasattr(new_class, "obj_model"):
                    raise ValueError(
                        "%s inherits from Relation and must therefore specify obj_model"
                        % name
                    )

                # `subj_model` or `obj_model` being a list was supported in an earlier
                # version of apis, but it is not anymore
                if isinstance(getattr(new_class, "subj_model", None), list):
                    raise ValueError("%s.subj_model must not be a list" % name)
                if isinstance(getattr(new_class, "obj_model", None), list):
                    raise ValueError("%s.obj_model mut not be a list" % name)

            if not new_class._meta.ordering:
                logger.warning(
                    f"{name} inherits from Relation but does not specify 'ordering' in its Meta class. "
                    "Empty ordering could result in inconsitent results with pagination. "
                    "Set a ordering or inherit the Meta class from Relation.",
                )

            return new_class


@functools.cache
def get_by_natural_key(natural_key: str):
    app_label, name = natural_key.lower().split(".")
    return ContentType.objects.get_by_natural_key(app_label, name).model_class()


class Relation(GenericModel, models.Model, metaclass=RelationModelBase):
    subj_content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, related_name="relation_subj_set"
    )
    subj_object_id = models.PositiveIntegerField(null=True)
    subj = GenericForeignKey("subj_content_type", "subj_object_id")
    obj_content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, related_name="relation_obj_set"
    )
    obj_object_id = models.PositiveIntegerField(null=True)
    obj = GenericForeignKey("obj_content_type", "obj_object_id")

    objects = RelationManager()

    class Meta:
        indexes = [
            models.Index(
                fields=["subj_content_type"], name="relations_r_subj_content_type"
            ),
            models.Index(fields=["subj_object_id"], name="relations_r_subj_object_id"),
            models.Index(
                fields=["obj_content_type"], name="relations_r_obj_content_type"
            ),
            models.Index(fields=["obj_object_id"], name="relations_r_obj_object_id"),
            models.Index(
                fields=["subj_content_type", "subj_object_id"],
                name="relations_r_subj_c_t_o_i",
            ),
            models.Index(
                fields=["obj_content_type", "obj_object_id"],
                name="relations_r_obj_c_t_o_i",
            ),
        ]

    def save(self, *args, **kwargs):
        subj_model = getattr(self, "subj_model", None)
        if subj_model and self.subj_content_type is subj_model:
            raise ValidationError(f"{self.subj} is not of type {subj_model}")
        obj_model = getattr(self, "obj_model", None)
        if obj_model and self.obj_content_type is obj_model:
            raise ValidationError(f"{self.obj} is not of type {obj_model}")
        super().save(*args, **kwargs)

    @property
    def subj_to_obj_text(self) -> str:
        if hasattr(self, "name"):
            return f"{self.subj} {self.name()} {self.obj}"
        return f"{self.subj} relation to {self.obj}"

    @property
    def obj_to_subj_text(self) -> str:
        if hasattr(self, "reverse_name"):
            return f"{self.obj} {self.reverse_name()} {self.subj}"
        return f"{self.obj} relation to {self.subj}"

    def __str__(self):
        return self.subj_to_obj_text

    @classmethod
    def subj_model_type(cls):
        model = cls.subj_model
        return get_by_natural_key(model) if isinstance(model, str) else model

    @classmethod
    def obj_model_type(cls):
        model = cls.obj_model
        return get_by_natural_key(model) if isinstance(model, str) else model

    @classmethod
    def name(cls) -> str:
        return cls._meta.verbose_name

    @classmethod
    def reverse_name(cls) -> str:
        return cls._meta.verbose_name + " reverse"

    @classmethod
    def name_and_reverse_name(cls) -> str:
        """
        Return a string with both the name and the reverse name.

        If they are identical, return only the name.
        """
        if cls.name() != cls.reverse_name():
            return f"{cls.name()} - {cls.reverse_name()}"
        return cls.name()



from django.forms import Form
from django.views.generic.detail import BaseDetailView, DetailView
from django.views.generic.edit import FormView
from django_tables2 import SingleTableMixin
from django_tables2.tables import table_factory

from apis_core.generic.helpers import first_member_match, module_paths
from apis_core.generic.views import (
    GenericModelMixin,
    GenericModelPermissionRequiredMixin,
)

from .tables import HistoryGenericTable


class HistoryView(
    GenericModelMixin, GenericModelPermissionRequiredMixin, SingleTableMixin, DetailView
):
    """
    This view lists the historical data of one object.
    The `GenericModel` this view acts upon is NOT the historical model, but the base model.
    I.e. it is not the `VersionPerson`, but the `Person`. Therefore the permission required
    to access this view refers to the `Person` model, NOT the `VersionPerson` model!
    """

    template_name = "history/history.html"
    permission_action_required = "view"

    def get_table_class(self):
        table_modules = module_paths(self.model, path="tables", suffix="HistoryTable")
        table_class = first_member_match(table_modules, HistoryGenericTable)
        return table_factory(self.model, table_class)

    def get_table_data(self):
        return self.get_object().get_history_data()


class HistoryReset(
    GenericModelMixin, GenericModelPermissionRequiredMixin, BaseDetailView, FormView
):
    """
    This view allows to reset a model instance to an earlier version.
    The `GenericModel` this view acts upon IS the historical model, so the permission
    required to access the view refers to the versioned model. For example for the
    `Person` model, this view acts upon the `VersionPerson` model and therefore the
    permission required is the `.versionperson_view` permission.
    """

    form_class = Form
    template_name = "history/reset.html"
    permission_action_required = "change"

    def form_valid(self, form):
        self.get_object().instance.save()
        return super().form_valid(form)

    def get_success_url(self):
        return self.get_object().instance.get_history_url()



import django_filters
from django.apps import apps
from django.conf import settings
from django.db import models
from django.forms import DateInput
from django.utils.encoding import force_str
from simple_history.utils import get_history_manager_for_model

from apis_core.generic.filtersets import GenericFilterSet
from apis_core.generic.helpers import default_search_fields, generate_search_filter


class ModelSearchFilter(django_filters.CharFilter):
    """
    This filter is a customized CharFilter that
    uses the `generate_search_filter` method to
    adapt the search filter to the model that is
    searched.
    It also extracts sets the help text based on
    the fields searched.
    """

    def __init__(self, *args, **kwargs):
        model = kwargs.pop("model", None)
        super().__init__(*args, **kwargs)

        if model is not None and "help_text" not in self.extra:
            field_names = [field.verbose_name for field in default_search_fields(model)]
            # use force_str on the fields verbose names to convert
            # lazy instances to string and join the results
            fields = ", ".join(map(force_str, field_names))
            self.extra["help_text"] = f"Search in fields: {fields}"

    def filter(self, qs, value):
        return qs.filter(generate_search_filter(qs.model, value))


def changed_since(queryset, name, value):
    history = get_history_manager_for_model(queryset.model)
    ids = history.filter(history_date__gt=value).values_list("id", flat=True)
    return queryset.filter(pk__in=ids)


class AbstractEntityFilterSet(GenericFilterSet):
    class Meta(GenericFilterSet.Meta):
        filter_overrides = {
            models.CharField: {
                "filter_class": django_filters.CharFilter,
                "extra": lambda f: {
                    "lookup_expr": "icontains",
                },
            },
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if model := getattr(self.Meta, "model", False):
            self.filters["search"] = ModelSearchFilter(model=model)
            self.filters.move_to_end("search", False)

        if "apis_core.history" in settings.INSTALLED_APPS:
            self.filters["changed_since"] = django_filters.DateFilter(
                label="Changed since",
                widget=DateInput(attrs={"type": "date"}),
                method=changed_since,
            )

        if apps.is_installed("apis_core.relations"):
            from apis_core.relations.filters import RelationFilter

            self.filters["relation"] = RelationFilter()



from collections import defaultdict

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.utils.encoding import force_str

from apis_core.apis_entities.utils import get_entity_content_types


class Datamodel:
    entities: [object]
    relations: [object] = []
    graph: dict() = {}

    def __init__(self):
        self.entities = get_entity_content_types()
        if "apis_core.relations" in settings.INSTALLED_APPS:
            from apis_core.relations.utils import relation_content_types

            self.relations = relation_content_types()
        self.make_graph()

    def edges(self):
        edges = defaultdict(list)
        for rel in self.relations:
            subj_class = rel.model_class().subj_model_type()
            obj_class = rel.model_class().obj_model_type()
            key = (
                ContentType.objects.get_for_model(subj_class).name,
                ContentType.objects.get_for_model(obj_class).name,
            )
            edges[key].append(force_str(rel.model_class().name()))
        return edges

    def make_graph(self):
        try:
            from pydot import Dot, Edge, Node

            graph = Dot(graph_type="digraph")
            graph.set_node_defaults(shape="record", rankdir="TB")
            graph.set_graph_defaults(nodesep="1", ranksep="1", TBbalance="max")
            for content_type in self.entities:
                model_class = content_type.model_class()
                node = Node(
                    content_type.name, label=force_str(model_class._meta.verbose_name)
                )
                node.set_URL(model_class.get_listview_url())
                node.set_fillcolor("#3399ff")
                node.set_style("filled")
                graph.add_node(node)
            for (subj, obj), names in self.edges().items():
                e = Edge(subj, obj, label="\n".join(names))
                graph.add_edge(e)
            self.graph = {"svg": graph.create_svg().decode(), "dot": graph.to_string()}
        except ImportError as e:
            self.graph["error"] = str(e)
        except FileNotFoundError as e:
            self.graph["error"] = "Please install graphviz - " + str(e)



from django.http import HttpResponseRedirect, QueryDict
from django.shortcuts import get_object_or_404
from drf_spectacular.utils import OpenApiParameter, OpenApiTypes, extend_schema
from rest_framework import viewsets
from rest_framework.response import Response

from apis_core.uris.models import Uri


class UriToObjectViewSet(viewsets.ViewSet):
    """
    This API route provides an endpoint for resolving URIs and forwarding
    them to the endpoint in the local instance. Pass a `uri` request
    parameter to resolve the uri.
    """

    @extend_schema(
        parameters=[
            OpenApiParameter(
                "uri", OpenApiTypes.URI, OpenApiParameter.QUERY
            ),  # path variable was overridden
        ],
        responses={301: None},
        description="This API route provides an endpoint for resolving URIs and forwarding them to the endpoint in the local instance. Pass a `uri` request parameter to resolve the uri.",
    )
    def list(self, request):
        params = request.query_params.dict()
        uri = params.pop("uri", None)
        if uri:
            u = get_object_or_404(Uri, uri=request.query_params.get("uri"))
            r = u.content_object.get_api_detail_endpoint()
            if params:
                r += "?" + QueryDict.from_keys(params).urlencode()
            return HttpResponseRedirect(r)
        return Response()



from django.urls import path
from django.views.generic import TemplateView

from apis_core.core.views import Dumpdata, PasswordChangeView

urlpatterns = [
    path("", TemplateView.as_view(template_name="base.html"), name="apis_index"),
    path("api/dumpdata", Dumpdata.as_view()),
    path(
        "profile/password-change/", PasswordChangeView.as_view(), name="password-change"
    ),
]



import inspect
from typing import Any

import django
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import AppRegistryNotReady
from django.db import models
from django.db.models import Q
from django.urls import reverse
from django.utils import timezone
from simple_history import utils
from simple_history.models import HistoricalRecords, ModelChange

from apis_core.generic.abc import GenericModel
from apis_core.generic.templatetags.generic import modeldict


class APISHistoricalRecords(HistoricalRecords):
    def get_m2m_fields_from_model(self, model):
        # Change the original simple history function to also return m2m fields
        m2m_fields = []
        try:
            for field in inspect.getmembers(model):
                if isinstance(
                    field[1],
                    django.db.models.fields.related_descriptors.ManyToManyDescriptor,
                ):
                    m2m_fields.append(getattr(model, field[0]).field)
        except AppRegistryNotReady:
            pass
        return m2m_fields

    def get_prev_record(self):
        """
        Get the previous history record for the instance. `None` if first.
        """
        history = utils.get_history_manager_from_history(self)
        return (
            history.filter(history_date__lt=self.history_date)
            .order_by("history_date")
            .first()
        )


class APISHistoryTableBase(GenericModel, models.Model):
    class Meta:
        abstract = True

    def get_absolute_url(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:generic:detail", args=[ct, self.history_id])

    def get_reset_url(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:history:reset", args=[ct, self.history_id])

    def get_diff(self, other_version=None):
        if self.history_type == "-":
            return None

        new_version_dict = modeldict(self.instance, exclude_noneditable=False)

        old_version = other_version or self.prev_record or None
        old_version_dict = dict()
        if old_version:
            old_version_dict = modeldict(
                old_version.instance, exclude_noneditable=False
            )

        # the `modeldict` method uses the field as a dict key but we are only interested
        # in the `.name` of the field, so lets replace the keys:
        new_version_dict = {key.name: value for key, value in new_version_dict.items()}
        old_version_dict = {key.name: value for key, value in old_version_dict.items()}

        newchanges = []
        for field, value in new_version_dict.items():
            old_value = old_version_dict.pop(field, None)
            if (value or old_value) and value != old_value:
                newchanges.append(ModelChange(field, old_value, value))
        for field, value in old_version_dict.items():
            if value is not None:
                newchanges.append(ModelChange(field, value, None))

        return sorted(newchanges, key=lambda change: change.field)


class VersionMixin(models.Model):
    history = APISHistoricalRecords(
        inherit=True,
        bases=[
            APISHistoryTableBase,
        ],
        custom_model_name=lambda x: f"Version{x}",
    )
    __history_date = None

    @property
    def _history_date(self):
        return self.__history_date or timezone.now()

    @_history_date.setter
    def _history_date(self, value):
        self.__history_date = value
        pass

    class Meta:
        abstract = True

    def get_history_url(self):
        ct = ContentType.objects.get_for_model(self)
        return reverse("apis_core:history:history", args=[ct, self.id])

    def _get_historical_relations(self):
        ret = set()
        if "apis_core.relations" in settings.INSTALLED_APPS:
            from apis_core.relations.utils import relation_content_types

            ct = ContentType.objects.get_for_model(self)

            rel_content_types = relation_content_types(any_model=type(self))
            rel_models = [ct.model_class() for ct in rel_content_types]
            rel_history_models = [
                model for model in rel_models if issubclass(model, VersionMixin)
            ]

            for model in rel_history_models:
                for historical_relation in model.history.filter(
                    Q(subj_object_id=self.id, subj_content_type=ct)
                    | Q(obj_object_id=self.id, obj_content_type=ct)
                ).order_by("history_id"):
                    ret.add(historical_relation)
        # If there is a newer version of a historical relation, also
        # add it to the set. This can be the case when a relation subject
        # or object was changed and the relation does not point to this
        # object anymore. We still want to show the change to make it
        # clear that the relation does not point to the object anymore.
        for historical_relation in ret.copy():
            if historical_relation.next_record:
                ret.add(historical_relation.next_record)
        return ret

    def get_history_data(self):
        data = []
        prev_entry = None
        queries = self._get_historical_relations()

        for entry in queries:
            if prev_entry is not None:
                if (
                    entry.history_date == prev_entry.history_date
                    and entry.history_user_id == prev_entry.history_user_id
                ):
                    entry.history_type = prev_entry.history_type
                    data[-1] = entry
                    prev_entry = entry
                    continue
            data.append(entry)
            prev_entry = entry
        data += [x for x in self.history.all()]
        data = sorted(data, key=lambda x: x.history_date, reverse=True)
        return data

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)



from django.apps import AppConfig


class MetainfoConfig(AppConfig):
    default_auto_field = "django.db.models.AutoField"
    name = "apis_core.uris"

    def ready(self):
        from . import signals  # noqa: F401



# Generated by Django 5.1 on 2024-09-25 07:18

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_metainfo", "0013_delete_collection"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="uri",
            name="domain",
        ),
        migrations.RemoveField(
            model_name="uri",
            name="loaded",
        ),
        migrations.RemoveField(
            model_name="uri",
            name="loaded_time",
        ),
        migrations.RemoveField(
            model_name="uri",
            name="rdf_link",
        ),
    ]



from django.utils.translation import gettext_lazy as _

from apis_core.apis_entities.abc import E21_Person, E53_Place, E74_Group
from apis_core.relations.models import Relation


class IsMemberOf(Relation):
    subj_model = E21_Person
    obj_model = E74_Group

    class Meta:
        abstract = True

    @classmethod
    def name(self) -> str:
        return _("is member of")

    @classmethod
    def reverse_name(self) -> str:
        return _("has as member")


class LivesIn(Relation):
    subj_model = E21_Person
    obj_model = E53_Place

    class Meta:
        abstract = True

    @classmethod
    def name(self) -> str:
        return _("lives in")

    @classmethod
    def reverse_name(self) -> str:
        return _("has habitant")


class IsLocatedIn(Relation):
    subj_model = E53_Place
    obj_model = E53_Place

    class Meta:
        abstract = True

    @classmethod
    def name(self) -> str:
        return _("is located in")

    @classmethod
    def reverse_name(self) -> str:
        return _("is location of")


class BornIn(Relation):
    subj_model = E21_Person
    obj_model = E53_Place

    class Meta:
        abstract = True

    @classmethod
    def name(self) -> str:
        return _("born in")

    @classmethod
    def reverse_name(self) -> str:
        return _("is birth place of")


class DiedIn(Relation):
    subj_model = E21_Person
    obj_model = E53_Place

    class Meta:
        abstract = True

    @classmethod
    def name(self) -> str:
        return _("died in")

    @classmethod
    def reverse_name(self) -> str:
        return _("is place of death of")



# Set up app with custom test models
# https://code.djangoproject.com/ticket/7835#comment:46
from django.apps import AppConfig, apps


def setup_test_app(package, label=None):
    """
    Setup a Django test app for the provided package to allow test models
    tables to be created if the containing app has migrations.

    This function should be called from app.tests __init__ module and pass
    along __package__.
    """
    app_config = AppConfig.create(package)
    app_config.apps = apps
    if label is None:
        containing_app_config = apps.get_containing_app_config(package)
        label = f"{containing_app_config.label}_tests"
    if label in apps.app_configs:
        raise ValueError(f"There's already an app registered with the '{label}' label.")
    app_config.label = label
    apps.app_configs[app_config.label] = app_config
    app_config.import_models()
    apps.clear_cache()


setup_test_app(__package__)



from rest_framework.generics import ListAPIView, RetrieveAPIView

from apis_core.history.serializers import (
    HistoryLogSerializer,
    HistoryObjectSerializer,
)


class EntityHistoryLogs(ListAPIView):
    serializer_class = HistoryLogSerializer

    def get_queryset(self):
        return (
            self.kwargs.get("contenttype")
            .model_class()
            .history.filter(id=self.kwargs.get("pk"))
        )


class GenericHistoryLog(RetrieveAPIView):
    serializer_class = HistoryObjectSerializer

    def get_queryset(self):
        return self.kwargs.get("contenttype").model_class().objects.all()



from django import template
from django.utils.safestring import mark_safe

from apis_core.utils.helpers import get_html_diff

register = template.Library()


@register.filter
def get_diff_old(change, shorten=0):
    return mark_safe(get_html_diff(a=change.old, b=change.new, show_b=False))


@register.filter
def get_diff_new(change, shorten=0):
    return mark_safe(get_html_diff(a=change.old, b=change.new, show_a=False))



from crum import get_current_request
from django.contrib import messages
from django.contrib.contenttypes.models import ContentType
from django.db.models.signals import post_save
from django.dispatch import receiver

from apis_core.collections.models import SkosCollection, SkosCollectionContentObject
from apis_core.history.models import APISHistoryTableBase


@receiver(post_save)
def add_to_session_collection(
    sender, instance, created, raw, using, update_fields, **kwargs
):
    """
    Add a created apis_core.history model instance to all the SkosCollections
    that are listed in the `session_collections` session variable.
    This needs the 'crum.CurrentRequestUserMiddleware' middleware to
    be enabled.
    """
    request = get_current_request()
    if isinstance(instance, APISHistoryTableBase) and request:
        for pk in request.session.get("session_collections", []):
            sc = SkosCollection.objects.get(pk=pk)
            content_type = ContentType.objects.get_for_model(instance)
            SkosCollectionContentObject.objects.create(
                collection=sc,
                content_type=content_type,
                object_id=instance.history_id,
            )
            messages.info(request, f"Tagged {instance} with tag {sc}")



import functools
import logging
import re

from django.conf import settings
from django.db.models.base import ModelBase
from django.urls import NoReverseMatch, reverse

from apis_core.apis_metainfo.models import RootObject
from apis_core.utils.settings import apis_base_uri

NEXT_PREV = getattr(settings, "APIS_NEXT_PREV", True)

logger = logging.getLogger(__name__)


class AbstractEntityModelBase(ModelBase):
    def __new__(metacls, name, bases, attrs):
        if name == "AbstractEntity":
            return super().__new__(metacls, name, bases, attrs)
        else:
            new_class = super().__new__(metacls, name, bases, attrs)
            if not new_class._meta.ordering:
                logger.warning(
                    f"{name} inherits from AbstractEntity but does not specify 'ordering' in its Meta class. "
                    "Empty ordering could result in inconsitent results with pagination. "
                    "Set a ordering or inherit the Meta class from AbstractEntity.",
                )

            return new_class


class AbstractEntity(RootObject, metaclass=AbstractEntityModelBase):
    """
    Abstract super class which encapsulates common logic between the
    different entity kinds and provides various methods relating to either
    all or one specific entity kind.

    Most of the class methods are designed to be used in the subclass as they
    are considering contexts which depend on the subclass entity type.
    So they are to be understood in that dynamic context.
    """

    class Meta:
        abstract = True

    @classmethod
    def get_or_create_uri(cls, uri):
        uri = str(uri)
        try:
            if re.match(r"^[0-9]*$", uri):
                p = cls.objects.get(pk=uri)
            else:
                p = cls.objects.get(uri__uri=uri)
            return p
        except Exception as e:
            print("Found no object corresponding to given uri." + e)
            return False

    # TODO
    @classmethod
    def get_entity_list_filter(cls):
        return None

    @functools.cached_property
    def get_prev_id(self):
        if NEXT_PREV:
            prev_instance = (
                type(self)
                .objects.filter(id__lt=self.id)
                .order_by("-id")
                .only("id")
                .first()
            )
            if prev_instance is not None:
                return prev_instance.id
        return False

    @functools.cached_property
    def get_next_id(self):
        if NEXT_PREV:
            next_instance = (
                type(self)
                .objects.filter(id__gt=self.id)
                .order_by("id")
                .only("id")
                .first()
            )
            if next_instance is not None:
                return next_instance.id
        return False

    def import_data(self, data):
        errors = super().import_data(data)
        if "same_as" in data:
            self._uris = data["same_as"]
            self.save()
        if "relations" in data:
            self.create_relations_to_uris = data["relations"]
            self.save()
        return errors

    def get_default_uri(self):
        try:
            route = reverse("GetEntityGenericRoot", kwargs={"pk": self.pk})
        except NoReverseMatch:
            route = reverse("apis_core:GetEntityGeneric", kwargs={"pk": self.pk})
        base = apis_base_uri().strip("/")
        return f"{base}{route}"



from django.contrib.contenttypes.models import ContentType
from rest_framework.reverse import reverse
from rest_framework.routers import APIRootView, DefaultRouter

from apis_core.generic.abc import GenericModel


class CustomAPIRootView(APIRootView):
    """
    The default basic root view for CustomDefaultRouter

    This view lists the output of the default APIRootView of the Django Rest Framework.
    In addition, it injects the routes of the genericmodelapi (those are not hardcoded
    registered but autogenerated based on the models that inherit from GenericModel).
    """

    def get(self, request, *args, **kwargs):
        response = super().get(request, *args, **kwargs)
        for content_type in ContentType.objects.all():
            if content_type.model_class() is not None and issubclass(
                content_type.model_class(), GenericModel
            ):
                route = "apis_core:generic:genericmodelapi-list"
                response.data[reverse(route, args=[content_type]).strip("/")] = reverse(
                    route,
                    args=[content_type],
                    kwargs=kwargs,
                    request=request,
                    format=kwargs.get("format"),
                )
        return response


class CustomDefaultRouter(DefaultRouter):
    """
    The CustomDefaultRouter only diverts from the Django Rest Framework DefaultRouter
    by setting the APIRootView to our CustomAPIRootView.
    """

    APIRootView = CustomAPIRootView



# Generated by Django 5.1.8 on 2025-06-04 05:22

import django.db.models.deletion
import simple_history.models
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("sample_project", "0008_remove_versiongroup_version_tag_and_more"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="VersionProfession",
            fields=[
                (
                    "id",
                    models.BigIntegerField(
                        auto_created=True, blank=True, db_index=True, verbose_name="ID"
                    ),
                ),
                ("name", models.CharField(blank=True, default="", max_length=1024)),
                ("history_id", models.AutoField(primary_key=True, serialize=False)),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical profession",
                "verbose_name_plural": "historical professions",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(
                simple_history.models.HistoricalChanges,
                models.Model,
            ),
        ),
    ]



"""
Copyright (c) 2023 Benoit Blanchon
License: MIT

This is copied from https://github.com/bblanchon/django-htmx-messages-framework
Commit a2da4176ac4a53fb0d7f42e8a1de398e4e5f9c1c
"""

import json

from django.contrib.messages import get_messages
from django.http import HttpRequest, HttpResponse
from django.utils.deprecation import MiddlewareMixin


class HtmxMessageMiddleware(MiddlewareMixin):
    """
    Middleware that moves messages into the HX-Trigger header when request is made with HTMX
    """

    def process_response(
        self, request: HttpRequest, response: HttpResponse
    ) -> HttpResponse:
        # The HX-Request header indicates that the request was made with HTMX
        if "HX-Request" not in request.headers:
            return response

        # Ignore redirections because HTMX cannot read the headers
        if 300 <= response.status_code < 400:
            return response

        # Extract the messages
        messages = [
            {"message": message.message, "tags": message.tags}
            for message in get_messages(request)
        ]
        if not messages:
            return response

        # Get the existing HX-Trigger that could have been defined by the view
        hx_trigger = response.headers.get("HX-Trigger")

        if hx_trigger is None:
            # If the HX-Trigger is not set, start with an empty object
            hx_trigger = {}
        elif hx_trigger.startswith("{"):
            # If the HX-Trigger uses the object syntax, parse the object
            hx_trigger = json.loads(hx_trigger)
        else:
            # If the HX-Trigger uses the string syntax, convert to the object syntax
            hx_trigger = {hx_trigger: True}

        # Add the messages array in the HX-Trigger object
        hx_trigger["messages"] = messages

        # Add or update the HX-Trigger
        response.headers["HX-Trigger"] = json.dumps(hx_trigger)

        return response



# Generated by Django 4.1.10 on 2023-09-06 07:04

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_metainfo", "0004_auto_20230310_0804"),
    ]

    operations = [
        migrations.DeleteModel(
            name="UriCandidate",
        ),
    ]



from django.db import models

from apis_core.apis_entities.abc import E21_Person, E53_Place, E74_Group
from apis_core.apis_entities.models import AbstractEntity
from apis_core.generic.abc import GenericModel
from apis_core.history.models import VersionMixin
from apis_core.relations.models import Relation


class Profession(VersionMixin, GenericModel, models.Model):
    name = models.CharField(blank=True, default="", max_length=1024)

    def __str__(self):
        return self.name


class Person(E21_Person, VersionMixin, AbstractEntity):
    profession = models.ManyToManyField(Profession, blank=True)


class Place(E53_Place, VersionMixin, AbstractEntity):
    pass


class Group(E74_Group, VersionMixin, AbstractEntity):
    pass


class IsCousinOf(Relation):
    subj_model = Person
    obj_model = Person

    @classmethod
    def reverse_name(self) -> str:
        return "is cousin of"


class IsPartOf(Relation):
    subj_model = Person
    obj_model = Group

    @classmethod
    def reverse_name(self) -> str:
        return "consists of"


class IsSiblingOf(Relation):
    subj_model = Person
    obj_model = Person

    @classmethod
    def reverse_name(self) -> str:
        return "is sibling of"


class LivesIn(Relation):
    subj_model = Person
    obj_model = Place

    @classmethod
    def reverse_name(self) -> str:
        return "has inhabitant"



from django.db import models

from apis_core.generic.abc import GenericModel


class Person(GenericModel, models.Model):
    first_name = models.CharField()
    last_name = models.CharField()

    def __str__(self):
        return f"{self.first_name} {self.last_name}"


class Dummy(GenericModel, models.Model):
    pass



# Generated by Django 4.2.10 on 2024-03-07 08:39

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_metainfo", "0009_remove_collection_collection_type"),
    ]

    operations = [
        migrations.RenameField(
            model_name="rootobject",
            old_name="name",
            new_name="deprecated_name",
        ),
    ]



# Generated by Django 4.2.8 on 2024-01-09 12:20

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("apis_entities", "0003_remove_tempentityclass_source"),
    ]

    operations = [
        migrations.DeleteModel(
            name="TempEntityClass",
        ),
    ]
